# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import scipy.fftpack
from scipy import signal
import numpy as np

def spectrum(x):
    """
    Return amplitude spectrum of signal.

    Parameters
    ----------
    x : array_like
        Signal.

    Returns
    -------
    : numpy.array
        Two-side amplitude spectrum.

    """
    return abs(scipy.fftpack.fft(x))

def expand_to(x, new_len):
    """
    Add zeros to signal. For doing magick with resolution in spectrum.

    Parameters
    ----------
    x : array_like
        Signal values.
    new_len : integer
        New length.

    Returns
    -------
    : numpy.array
        Signal expanded by zeros.
    
    """
    if new_len <= len(x):
        return x
    x_exp = np.zeros(new_len)
    x_exp[0 : len(x)] = x
    return x_exp

def stft(x, dt, nseg, nstep, window='hanning', nfft=None, padded=False):
    """
    Return result of short-time fourier transform.

    Parameters
    ----------
    x : numpy.ndarray
        Signal.
    dt : float 
       Sampling period.
    window : str
        Type of window.
    nseg : int
        Length of segment (in samples).
    nstep : int
        Length of step (in samples).
    nfft : int 
        Length of the FFT. If None or less than nseg, the FFT length is nseg.

    Returns
    -------

    : list of numpy.ndarray
        Result of STFT.
    
    """
    wind = signal.get_window(window, nseg)
    Xs=[]
    if padded:
        L = len(x) + (nseg - len(x) % nseg) % nseg
        x = expand_to(x, L)

    if not nfft:
        nseg_exp = nseg
    else:
        nseg_exp = max(nseg, nfft)
        
    for i in range(0, len(x)-nseg + 1, nstep):
        seg = x[i : i+nseg] * wind
        seg = expand_to(seg, nseg_exp)
        X = spectrum(seg)
        Xs.append(X)
    return Xs



# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

def read_signal_from_csv(file_name, t_column=0, x_column=1, start_line=1, fs=None, delimiter=';'):
    """ 
    Read signal from CSV-file

    Parameters
    ----------
    file_name : str
        Name of input file
    t_column : integer
        Number of column with time values
    x_column : integer
        Number of column with signal values
    start_line : integer
        Number of first line with sample
    fs : Sampling frequency
        Sampling frequency. If used t_column is ignored
    delimiter : str
        Delimiter used in CSV-file

    Returns
    -------
    x : np.array
        Values of signal
    x : np.array
        Values of time

    """
    data = np.genfromtxt(file_name, delimiter=delimiter)
    data = data.transpose()
    x = np.array(data[x_column][start_line:])
    if not fs:
        t = np.array(data[t_column][start_line:])
    else:
        t = np.linspace(0, (len(x)-1)/fs, len(x))
    return x, t



# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from scipy.signal import butter, lfilter
from scipy.fftpack import fft
import numpy as np

def butter_lowpass(cutoff, fs, order):
    """ 
    Calculate a and b coefficients for Butterworth lowpass filter

    Parameters
    ----------
    cutoff : float
        Cutoff frequency (Hz)
    fs : float
        Sampling frequency (Hz)
    order : integer
        Order of filter

    Returns
    -------
    b : np.array
        b-values
    a : np.array
        a-values

    """
    nyq = 0.5 * fs
    b, a = butter(order, cutoff/nyq, btype='low')
    return b, a

def butter_lowpass_filter(x, cutoff, fs, order):
    """ 
    Filter signal with Butterworth lowpass filter

    Parameters
    ----------
    x : array_like
        Signal values
    cutoff : float
        Cutoff frequency (Hz)
    fs : float
        Sampling frequency (Hz)
    order : integer
        Order of filter

    Returns
    -------
    x_new : np.array
        Values of filtered signal

    """
    b, a = butter_lowpass(cutoff, fs, order)
    return lfilter(b, a, x)

def butter_bandpass(lowcut, highcut, fs, order):
    """ 
    Calculate a and b coefficients for Butterworth bandpass filter

    Parameters
    ----------
    lowcut : float
        Low-cut frequency (Hz)
    highcut : float
        High-cut frequency (Hz)
    fs : float
        Sampling frequency (Hz)
    order : integer
        Order of filter

    Returns
    -------
    b : np.array
        b-values
    a : np.array
        a-values

    """
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a

def butter_bandpass_filter(x, lowcut, highcut, fs, order):
    """ 
    Filter signal with Butterworth bandpass filter

    Parameters
    ----------
    x : array_like
        Signal values
    lowcut : float
        Low-cut frequency (Hz)
    highcut : float
        High-cut frequency (Hz)
    fs : float
        Sampling frequency (Hz)
    order : integer
        Order of filter

    Returns
    -------
    x_new : np.array
        Values of filtered signal

    """
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    return lfilter(b, a, x)

def find_butt_bandpass_order(band, fs):
    """
    Claculate the order of Butterworth bandpass filter using minimization of metric between ideal and real frequency response.

    Paramaters
    ----------
    band : array_like
        Pair of frequencies. Bounds of bandpass (Hz)
    fs : float
        Sampling rate (Hz)

    Returns
    -------
    order : integer
        Order of filter

    """
    # TODO: Write a set of functions or stubs, all about order
    N = round(60 * 120 * fs)
    unit_pulse = np.zeros(N)
    unit_pulse[1] = 1
    n1 = round((band[0] * N//2) / (fs/2))
    n2 = round((band[1] * N//2) / (fs/2))
    ideal_fr = np.zeros(N)
    f = np.fft.fftfreq(N, 1/fs)
    ind = (f>= band[0])&(f<= band[1])
    ideal_fr[ind] = 1
    ideal_fr = ideal_fr[:N//2]
    prev_metric = np.inf
    for order in range(3, 21):
        impulse_response = butter_bandpass_filter(unit_pulse, band[0], band[1], fs, order)
        fr = abs(fft(impulse_response))[:N//2]
        metric = np.sum((fr - ideal_fr)**2)**0.5
        best_order = order
        if (np.isnan(metric)) or (metric >= prev_metric):
            best_order -= 1
            break
        prev_metric = metric
    return best_order-1 # TODO: Think about it

def haar_one_step(x, t, denominator=2):
    """ 
    One cascade of Haar transform.

    Parameters
    ----------
    x : array_like
        Signal values
    t : array_like
        Time values
    denominator : integer
        Denominator used in Haar transform (default is 2)

    Returns
    -------
    x_s : np.array
        Scaled signal values
    x_d : np.array
        Details of x
    t_new : np.array
        Decimated time values

    """
    # TODO: use t or fs in arguments
    x_s = []
    x_d = []
    t_new = []
    for x_left, x_right in zip(x[::2], x[1::2]):
        x_s.append((x_left+x_right)/denominator)
        x_d.append((x_left-x_right)/denominator)
    t_new = t[1::2]
    return np.array(x_s), np.array(x_d), np.array(t_new)

def haar_scaling(x, t, steps_number):
    """ 
    Scaling with Haar transform.

    Parameters
    ----------
    x : array_like
        Signal values
    t : array_like
        Time values
    steps_number : integer
        Number of cascades

    Returns
    -------
    x_s : np.array
        Scaled signal values
    t_new : np.array
        Decimated time values

    """
    # TODO: validation of steps_number
    x_s = x.copy()
    t_new = t.copy()
    for i in range(steps_number):
        x_s, x_d, t_new = haar_one_step(x_s, t_new, denominator=2)
    return x_s, t_new



# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the Lesser GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Lesser GNU General Public License for more details.

# You should have received a copy of the Lesser GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

def freq_stupid(x, fs):
    """
    Calculate frequency of oscillating signal by extremums

    Parameters
    ----------
    x : array_like
        Values of input signals
    fs : float
        Sampling frequency (Hz)

    Returns
    -------
    freq : float
        Frequency

    """
    T = len(x)/fs
    n_max = 0
    n_min = 0
    for x_prev, x_current, x_next in zip(x[:-2], x[1:-1], x[2:]):
        if (x_prev < x_current) and (x_current >= x_next):
            n_max += 1
        if (x_prev > x_current) and (x_current <= x_next):
            n_min += 1
    n = (n_max + n_min) / 2
    return n / T

def freqs_stupid(x, fs, window_width = 1024, window_step = 512):
    """
    Calculate an array of frequencies of oscillating signal using window

    Parameters
    ----------
    x : array_like
        Values of input signals
    fs : float
        Sampling frequency (Hz)
    window_width : integer
        Width of window (Samples)
    window_step : integer
        Distance between centers of nearby windows (Samples)

    Returns
    -------
    freqs : np.array
        Frequency values
    t_new : time values

    """
    freqs = []
    t_new = []
    start = 0
    stop = window_width
    if stop > len(x):
        return freqs, t_new
    while True:
        f = freq_stupid(x[start : stop], fs)
        freqs.append(f)
        t_new.append((stop-1)/fs)
        start += window_step
        stop += window_step
        if stop > len(x):
            break
    return np.array(freqs), np.array(t_new)
